from os import path

from ziplineio.response import StaticFileResponse


def _get_headers(filename: str) -> dict[str, str]:
    if filename.endswith(".css"):
        return {"Content-Type": "text/css"}
    elif filename.endswith(".js"):
        return {"Content-Type": "application/javascript"}
    elif filename.endswith(".html"):
        return {"Content-Type": "text/html"}
    return {"Content-Type": "text/plain"}


def staticfiles(filepath: str, path_prefix: str):
    async def handler(req, ctx):
        print(f"Request path: {req.path}")
        if req.path.startswith(path_prefix):
            # remove path prefix
            _filepath = path.join(filepath, req.path[len(path_prefix) :])
            # add filepath
            _filepath = filepath + _filepath
            # get full path
            _filepath = path.abspath(_filepath)
            headers = _get_headers(_filepath)
            return StaticFileResponse(_filepath, headers)
        return req, ctx

    return handler



from functools import wraps
from typing import Any

from ziplineio.handler import Handler


class MemoryCache:
    _instance = None
    _cache: dict[str, str]

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(cls, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        # Ensure the cache is only initialized once
        if not hasattr(self, "_cache"):
            self._cache = {}

    def get(self, key: str) -> str | None:
        return self._cache.get(key, None)

    def set(self, key: str, value: str) -> None:
        self._cache[key] = value

    def __call__(self, handler: Handler) -> Any:
        @wraps(handler)
        def wrapper(*args, **kwargs):
            # Create a key based on the arguments
            key = (args, frozenset(kwargs.items()))

            if key not in self._cache:
                # Call the function and store the result in the cache
                secache[key] = func(*args, **kwargs)
            return cache[key]

        return wrapper


memory_cache1 = MemoryCache()
memory_cache2 = MemoryCache()

print(memory_cache1 is memory_cache2)



import unittest
from ziplineio import utils


class TestUtils(unittest.TestCase):
    def test_clean_url(self):
        url = "/user/:id"
        parsed = utils.clean_url(url)

        # Assertions
        self.assertEqual(parsed, "/user/*")

    def test_match_url_pattern(self):
        pattern = "/user/:id"
        url = "/user/123"
        parsed = utils.match_url_pattern(pattern, url)

        # Assertions
        self.assertEqual(parsed, {"id": "123"})



from typing import Any, Callable
from ziplineio.request import Request


# Injected services is a dictionary that stores the services that are injected.
# Services are stored in the dictionary based on the scope of the service.
# Default, handler-level services are stored in the 'func' scope.
# App-level services are stored in the 'app' scope.
# Router level services are stored in the scope of router's id.

# ***
# {
#    "func": {},
#    "app": {},
#    "<some-uuid>": {}
# }


class DependencyInjector:
    _injected_services: dict[str, Any]

    def __init__(self) -> None:
        self._injected_services = {
            "app": {},
            "func": {},
        }

    def inject(
        self, service_class: Any, name: str = None, scope: str = "func"
    ) -> Callable[[Callable], Callable]:
        def decorator(handler: Callable | None) -> Callable:
            instance, service_name = self.add_injected_service(
                service_class, name, scope
            )

            async def wrapped_handler(req: Request, **kwargs):
                return await handler(req, **kwargs, **{service_name: instance})

            return wrapped_handler

        return decorator

    def add_injected_service(
        self, service_class: Any, name: str = None, scope: str = "func"
    ) -> None:
        service_name = name if name else service_class.__name__.lower()

        if scope not in self._injected_services:
            self._injected_services[scope] = {}

        services = self._injected_services[scope]

        if service_name in services:
            instance = services[service_name]
        elif isinstance(service_class, type):
            instance = service_class()
        else:
            instance = service_class

        services[service_name] = instance
        return instance, service_name

    def get_injected_services(self, scope: str = "func") -> dict[str, Any]:
        if scope not in self._injected_services:
            self._injected_services[scope] = {}
        return self._injected_services.get(scope, {})


injector = DependencyInjector()


# function to inject a service into a handler outside the main file
# (services, etc)
def inject(
    service_class: Any, name: str = None, scope: str = "func"
) -> Callable[[Callable], Callable]:
    return injector.inject(service_class, name, scope)



from multiprocessing import Process

import unittest


import uvicorn
import requests
import asyncio

from ziplineio.app import App


app = App()
app.static("test/mocks/static", path_prefix="/static")


@app.get("/")
async def home(req):
    return {"message": "Hello, world!"}


def run_server():
    uvicorn.run(app, port=5050)


class TestResponse(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        """Bring server up."""
        self.proc = Process(target=run_server, args=(), daemon=False)
        self.proc.start()
        await asyncio.sleep(0.2)  # time for the server to start

    async def test_static_file(self):
        r = requests.get("http://localhost:5050/static/css/test.css")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.headers["Content-Type"], "text/css")
        self.assertTrue("background-color: #f0f0f0;" in r.text)
        print(r.content)

    async def asyncTearDown(self):
        self.proc.terminate()



import json
from typing import List, Tuple, TypedDict, Dict


from ziplineio.exception import BaseHttpException


class RawResponse(TypedDict):
    headers: List[Tuple[bytes, bytes]]
    status: int
    body: bytes


class Response:
    def __init__(self, status: int, headers: Dict[str, str], body: str):
        self.status = status
        self.headers = headers
        self.body = body

    status: int
    headers: Dict[str, str]
    body: str

    def __len__(self) -> int:
        return 1


class StaticFileResponse(Response):
    def __init__(self, file_path: str, headers: Dict[str, str]):
        body = self.get_file(file_path)
        super().__init__(200, headers, body)

    def get_file(self, file_path: str) -> bytes:
        with open(file_path, "rb") as file:
            return file.read()


class JinjaResponse(Response):
    def __init__(self, body: str):
        super().__init__(200, {"Content-Type": "text/html"}, body)


def format_headers(headers: Dict[str, str] | None) -> List[Tuple[bytes, bytes]]:
    if headers is None:
        return []
    return [(bytes(k, "utf-8"), bytes(v, "utf-8")) for k, v in headers.items()]


def format_body(body: bytes | str | dict) -> bytes:
    if isinstance(body, bytes):
        return body
    elif isinstance(body, str):
        return bytes(body, "utf-8")
    elif isinstance(body, dict):
        return bytes(json.dumps(body), "utf-8")
    raise ValueError("Invalid body type")


def format_response(
    response: bytes | dict | str | Response, default_headers: dict[str, str]
) -> RawResponse:
    print("TYUUPE:")
    print(type(response))
    if isinstance(response, bytes):
        raw_response = {
            "headers": [
                (b"content-type", b"text/plain"),
            ],
            "status": 200,
            "body": response,
        }
    elif isinstance(response, str):
        raw_response = {
            "headers": [
                (b"content-type", b"text/plain"),
            ],
            "status": 200,
            "body": bytes(response, "utf-8"),
        }
    elif isinstance(response, dict):
        raw_response = {
            "headers": [
                (b"content-type", b"application/json"),
            ],
            "status": 200,
            "body": bytes(json.dumps(response), "utf-8"),
        }
    elif isinstance(response, Response):
        print("HRE")
        raw_response = {
            "headers": format_headers(response.headers),
            "status": response.status,
            "body": format_body(response.body),
        }
    elif isinstance(response, BaseHttpException):
        raw_response = {
            "headers": [
                (b"content-type", b"application/json"),
            ],
            "status": response.status_code,
            "body": format_body(response.message),
        }
    else:
        print(response)
        raise ValueError("Invalid response type")

    raw_response["headers"].extend(format_headers(default_headers))
    return raw_response



from multiprocessing import Process
import asyncio
import requests
import uvicorn
import unittest
import unittest.async_case


from ziplineio.app import App
from ziplineio.router import Router


app = App()


@app.get("/")
async def handler(req):
    return {"message": "Hello, world!"}


# Will be made multithreaded
@app.get("/sync-thread")
def sync_handler(req):
    return {"message": "Hello, sync world!"}


user_router = Router("/user")


@user_router.get("/:id")
async def user_handler(req):
    return {"message": f"User {req.path_params['id']} received"}


def run_server():
    uvicorn.run(app, port=5050)


class TestE2E(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        """Bring server up."""
        self.proc = Process(target=run_server, args=(), daemon=False)
        self.proc.start()
        await asyncio.sleep(0.2)  # time for the server to start

    async def test_basic_route(self):
        response = requests.get("http://localhost:5050/")
        print(response.json())
        self.assertEqual(response.status_code, 200)

    async def test_sync_route(self):
        response = requests.get("http://localhost:5050/sync-thread")
        print(response.content)
        self.assertEqual(response.status_code, 200)

    async def asyncTearDown(self):
        self.proc.terminate()



import re
import inspect
import asyncio
from typing import Dict
from ziplineio.request import Request
from ziplineio.response import RawResponse, format_response
from ziplineio.handler import Handler
from ziplineio.models import ASGIScope
from ziplineio.exception import BaseHttpException
from ziplineio import settings


async def call_handler(
    handler: Handler, req: Request, format: bool = True
) -> RawResponse:
    try:
        if not inspect.iscoroutinefunction(handler):
            response = await asyncio.to_thread(handler, req)
        else:
            response = await handler(req)

    except BaseHttpException as e:
        response = e
    except Exception as e:
        print(e)
        response = BaseHttpException(e, 500)

    if format:
        return format_response(response, settings.DEFAULT_HEADERS)
    else:
        return response


def parse_scope(scope: ASGIScope) -> Request:
    query_string = scope["query_string"].decode("utf-8")

    print(f"query_string: {query_string}")

    if query_string == "":
        query_params = {}
    else:
        query_params = dict(qp.split("=") for qp in query_string.split("&"))

    split_path = scope["path"].split(":")
    if len(split_path) > 1:
        path_params = dict(zip(split_path[1:], scope["path"].split("/")[1:]))
    else:
        path_params = {}

    headers = dict((k.decode("utf-8"), v.decode("utf-8")) for k, v in scope["headers"])
    return Request(
        method=scope["method"],
        path=scope["path"],
        query_params=query_params,
        path_params=path_params,
        headers=headers,
        body="",
    )


def clean_url(path: str) -> str:
    return re.sub(r":\w+", "*", path)


def match_url_pattern(pattern: str, url: str) -> Dict[str, str]:
    """
    Match a URL against a pattern with wildcards and return a dictionary of captured parameters.

    Args:
        pattern (str): The pattern with wildcards (e.g., 'user/:id').
        url (str): The URL to match (e.g., 'users/123').

    Returns:
        Dict[str, str]: A dictionary of captured parameters (e.g., {'id': '123'}).
    """
    # Convert the pattern into a regular expression
    # Replace ':param' with '([^/]+)' to capture the parameter value
    regex_pattern = re.sub(r":(\w+)", r"(?P<\1>[^/]+)", pattern)
    regex_pattern = "^" + regex_pattern + "$"

    # Compile the regular expression
    pattern_re = re.compile(regex_pattern)

    # Match the URL against the pattern
    match = pattern_re.match(url)
    if not match:
        return {}

    # Return the captured parameters as a dictionary
    return match.groupdict()



from typing import Any

from ziplineio.response import JinjaResponse
from ziplineio.utils import call_handler


def jinja(env: Any, template_name: str):
    template = env.get_template(template_name)

    def decorator(handler):
        async def wrapped_handler(*args, **kwargs):
            context = await call_handler(handler, *args, **kwargs, format=False)
            rendered = template.render(context)
            return JinjaResponse(rendered)

        return wrapped_handler

    return decorator



from typing import TypedDict, List, Tuple, Optional, Dict, Any, Protocol


class ASGISend(Protocol):
    async def __call__(self, message: Dict[str, Any]) -> None:
        pass


class ASGIScope(TypedDict):
    type: str  # Typically 'http' or 'websocket'
    http_version: str  # '1.1' for HTTP/1.1, '2' for HTTP/2
    method: str  # HTTP method (e.g., 'GET', 'POST')
    scheme: str  # 'http' or 'https'
    path: str  # URL path (e.g., '/hello')
    raw_path: bytes  # The raw path as bytes
    query_string: bytes  # Raw query string in bytes (e.g., b'name=John')
    headers: List[Tuple[bytes, bytes]]  # List of (header_name, header_value) as bytes
    client: Optional[Tuple[str, int]]  # Client address (IP, port)
    server: Optional[Tuple[str, int]]  # Server address (IP, port)
    root_path: str  # The root path for routing
    app: dict  # App-specific data, typically empty
    state: dict  # State shared across different parts of the application
    extensions: dict  # ASGI extensions in use (if any)



import asyncio
from multiprocessing import Process
import unittest

import requests
import uvicorn
from jinja2 import Environment, PackageLoader, select_autoescape
from ziplineio.app import App

from ziplineio.html.jinja import jinja

env = Environment(loader=PackageLoader("test.mocks"), autoescape=select_autoescape())

app = App()


@app.get("/")
@jinja(env, "home.html")
def home(req):
    return {"message": "Hello, world!"}


def run_server():
    uvicorn.run(app, port=5050)


class TestHtml(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        """Bring server up."""
        self.proc = Process(target=run_server, args=(), daemon=False)
        self.proc.start()
        await asyncio.sleep(0.2)  # time for the server to start

    async def test_render_jinja(self):
        response = requests.get("http://localhost:5050/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], "text/html")
        self.assertTrue("<p>Welcome to the home page!</p>" in response.text)

    async def asyncTearDown(self):
        self.proc.terminate()



import re
from typing import List, Callable, Tuple
from ziplineio import response
from ziplineio.handler import Handler
from ziplineio.request import Request
from ziplineio.response import Response


def middleware(middlewares: List[Callable]) -> Callable[[Callable], Callable]:
    def decorator(handler: Callable) -> Callable:
        async def wrapped_handler(req: Request, **kwargs):
            # Ensure 'ctx' is always present in kwargs

            kwargs.setdefault("ctx", {})
            # Run the middleware stack
            try:
                req, kwargs, res = await run_middleware_stack(middlewares, req, kwargs)
            except Exception as e:
                return {
                    "status": 500,
                    "headers": [],
                    "body": b"Internal server error: " + str(e).encode(),
                }

            if res is not None:
                return res
            else:
                return await handler(req, **kwargs)

        return wrapped_handler

    return decorator


async def run_middleware_stack(
    middlewares: list[Handler], request: Request, kwargs
) -> Tuple[Request, dict, bytes | str | dict | Response | None]:
    for middleware in middlewares:
        # if the middleware func takes params, pass them in. Otherwise, just pass req

        if "ctx" not in kwargs:
            kwargs["ctx"] = {}

        print("KWARGS")
        print(kwargs)
        print(middleware.__code__.co_varnames)
        if len(middleware.__code__.co_varnames) > 1:
            _res = await middleware(request, **kwargs)
        else:
            _res = await middleware(request)

        print("RES")
        print(_res)

        if len(_res) != 2:
            req = _res
            middleware_ctx = kwargs["ctx"]
        else:
            req, middleware_ctx = _res
            kwargs["ctx"].update(middleware_ctx)

        if not isinstance(req, Request):
            response = req
            return request, kwargs, response

    return request, kwargs, None
